import glob
from importlib import import_module
from pathlib import Path
from typing import Callable, final

from stelvio.aws.function import get_project_root
from stelvio.component import ComponentRegistry


@final
class StelvioApp:
    def __init__(
        self,
        name: str,
        modules: list[str],
        link_configs: dict[type, Callable] | None = None,
    ):
        self.name = name
        self._modules = modules
        if link_configs:
            for component_type, fn in link_configs.items():
                self.set_default_link_for(component_type, fn)

    @staticmethod
    def set_default_link_for(component_type, func):
        ComponentRegistry.register_link_config_creator(component_type, func)

    def run(self):
        self.drive()

    def drive(self):
        self._load_modules(self._modules, get_project_root())
        # Vroooom through those infrastructure deployments
        # like an Alfa through those hairpins
        for i in ComponentRegistry.all_instances():
            print(i.name)
            i._ensure_resource()

    def _load_modules(self, modules: list[str], project_root: Path):
        exclude_dirs = {"__pycache__", "build", "dist", "node_modules", ".egg-info"}

        for pattern in modules:
            # Direct module import
            if "." in pattern and not any(c in pattern for c in "/*?[]"):
                import_module(pattern)
                continue

            # Glob pattern
            files = glob.glob(pattern, root_dir=project_root, recursive=True)

            for file in files:
                path = Path(file)

                # Skip hidden folders (any part starts with .)
                if any(part.startswith(".") for part in path.parts):
                    continue

                if path.suffix == ".py" and not any(
                    excluded in path.parts for excluded in exclude_dirs
                ):

                    parts = list(path.with_suffix("").parts)
                    if all(part.isidentifier() for part in parts):
                        module_path = ".".join(parts)
                        import_module(module_path)



import re
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from functools import cache
from typing import Unpack, final, Literal, TypeAlias, Tuple

from pulumi import StringAsset, ResourceOptions
from pulumi_aws import get_region, get_caller_identity
from pulumi_aws.iam import (
    Role,
    get_policy_document,
    GetPolicyDocumentStatementArgs,
    GetPolicyDocumentStatementPrincipalArgs,
    RolePolicyAttachment,
)
from pulumi_aws.lambda_ import Permission

HTTP_METHODS = Literal["GET", "POST", "PUT", "PATCH", "DELETE", "HEAD", "OPTIONS"]

import pulumi
from pulumi_aws.apigateway import Resource, Method, Integration, Account

from stelvio.aws.function import (
    FunctionConfigDict,
    FunctionConfig,
    Function,
    FunctionAssetsRegistry,
)
from stelvio.component import Component, ComponentRegistry
from pulumi_aws.apigateway import RestApi, Deployment, Stage
from hashlib import sha256
import json

from enum import Enum

API_GATEWAY_LOGS_POLICY = (
    "arn:aws:iam::aws:policy/service-role/AmazonAPIGatewayPushToCloudWatchLogs"
)


# These are methods supported by api gateway
class HTTPMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    ANY = "ANY"


HTTPMethodLiteral = Literal[
    "GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "OPTIONS", "ANY", "*"
]

HTTPMethodInput: TypeAlias = (
    str | HTTPMethodLiteral | HTTPMethod | list[str | HTTPMethodLiteral | HTTPMethod]
)


def _validate_single_method(method: str | HTTPMethod) -> None:
    # Convert to string if it's enum
    if isinstance(method, HTTPMethod):
        method = method.value
    method_upper_case = method.upper()
    # Handle ANY and * as synonyms
    if method_upper_case in ("ANY", "*"):
        return

    # Check against enum values
    valid_methods = {m.value for m in HTTPMethod if m != HTTPMethod.ANY}
    if method_upper_case not in valid_methods:
        raise ValueError(f"Invalid HTTP method: {method}")


def normalize_method(method: str | HTTPMethodLiteral | HTTPMethod) -> str:
    if isinstance(method, str):
        return method.upper() if method != "*" else HTTPMethod.ANY.value
    if isinstance(method, HTTPMethod):
        return method.value


@final
@dataclass(frozen=True)
class _ApiRoute:
    method: HTTPMethodInput
    path: str
    handler: FunctionConfig | FunctionConfigDict | Function

    def __post_init__(self):
        # https://docs.aws.amazon.com/apigateway/latest/developerguide/limits.html
        # TODO: validate handler type
        self._validate_path()
        self._validate_method()

    def _validate_path(self):
        if not self.path.startswith("/"):
            raise ValueError("Path must start with '/'")
        # This includes query params and it's bit higher for regional apis
        if len(self.path) > 8192:  # example limit
            raise ValueError("Path too long")
        # Empty braces check
        if "{}" in self.path:
            raise ValueError("Empty path parameters not allowed")
        # Find all parameters
        params = re.findall(r"{([^}]+)}", self.path)
        # Check max number of parameters (AWS limit)
        if len(params) > 10:
            raise ValueError("Maximum of 10 path parameters allowed")
        # Adjacent parameters check
        if re.search(r"}{", self.path):
            raise ValueError("Adjacent path parameters not allowed")
        # Duplicate check
        if len(params) != len(set(params)):
            raise ValueError("Duplicate path parameters not allowed")
        for param in params:
            if param.endswith("+"):
                if param != "proxy+":
                    raise ValueError("Only {proxy+} is supported for greedy paths")
                if self.path.index(f"{{{param}}}") != len(self.path) - len(
                    f"{{{param}}}"
                ):
                    raise ValueError("Greedy parameter must be at the end of the path")
                continue

            if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", param):
                raise ValueError(f"Invalid parameter name: {param}")

    # TODO: allow get Get, etc.
    def _validate_method(self):
        if isinstance(self.method, (str, HTTPMethod)):
            _validate_single_method(self.method)
        elif isinstance(self.method, list):
            if not self.method:  # empty check
                raise ValueError("Method list cannot be empty")
            for m in self.method:
                if not isinstance(m, (str, HTTPMethod)):
                    raise ValueError(f"Invalid method type in list: {type(m)}")
                if isinstance(m, HTTPMethod) and m == HTTPMethod.ANY:
                    raise ValueError("ANY not allowed in method list")
                if isinstance(m, str) and m in ("ANY", "*"):
                    raise ValueError("ANY and * not allowed in method list")
                _validate_single_method(m)
        else:
            raise ValueError(
                f"Method must be string, HTTPMethod, or list of them, got {type(self.method)}"
            )

    @property
    def methods(self) -> list[str]:
        if isinstance(self.method, list):
            return [normalize_method(m) for m in self.method]
        else:
            return [normalize_method(self.method)]


class Api(Component):
    _routes: list[_ApiRoute]

    def __init__(self, name: str) -> None:
        self._routes = []
        super().__init__(name)

    def route(
        self,
        http_method: HTTPMethodInput,
        path: str,
        handler: str | FunctionConfig | FunctionConfigDict | Function | None = None,
        /,
        **opts: Unpack[FunctionConfigDict],
    ) -> None:
        """Add a route to the API.

        The route handler can be specified in three ways:
        1. As a complete configuration object (FunctionConfig, FunctionConfigDict, or Function)
        2. As a handler path string with optional FunctionConfigDict fields as kwargs
        3. As FunctionConfigDict fields passed directly as keyword arguments

        Args:
            http_method: HTTP method(s) to handle. Can be:
                - String ("GET", "POST", etc.)
                - HTTPMethod enum value
                - List of methods for multiple method support
                - "ANY" or "*" to handle all methods
            path: URL path for the route
            handler: Route handler specification. Can be:
                - Function handler path as string
                - Complete FunctionConfig object
                - FunctionConfigDict dictionary
                - Function instance
                - None (if handler is specified in opts)
            **opts: Additional FunctionConfigDict fields when using handler path

        Raises:
            ValueError: If the configuration is ambiguous or incomplete
            TypeError: If handler is of invalid type

        Examples:
            # Single method
            api.route("GET", "/users", "users.list", memory_size=128)
            api.route(HTTPMethod.GET, "/users", "users.list")

            # Multiple methods
            api.route(["GET", "POST"], "/users", "users.handle")

            # All methods
            api.route("ANY", "/users", "users.handle")

            # Configuration examples
            api.route("GET", "/users", {"handler": "users.list", "memory_size": 128})
            api.route("GET", "/users", handler="users.list", memory_size=128)
        """
        api_route = self._create_route(http_method, path, handler, opts)
        self._routes.append(api_route)

    @staticmethod
    def _create_route(
        http_method: HTTPMethodInput,
        path: str,
        handler: str | FunctionConfig | FunctionConfigDict | Function | None,
        opts: dict,
    ) -> _ApiRoute:
        if isinstance(handler, (dict, FunctionConfig, Function)):
            if opts:
                raise ValueError(
                    "Invalid configuration: cannot combine complete handler "
                    "configuration with additional options"
                )
            return _ApiRoute(http_method, path, handler)

        if isinstance(handler, str):
            if opts and "handler" in opts:
                raise ValueError(
                    "Ambiguous handler configuration: handler is specified both as "
                    "positional argument and in options"
                )
            return _ApiRoute(http_method, path, FunctionConfig(handler=handler, **opts))

        if handler is None:
            if not opts or "handler" not in opts:
                raise ValueError(
                    "Missing handler configuration: when handler argument is None, "
                    "'handler' option must be provided"
                )
            return _ApiRoute(http_method, path, FunctionConfig(**opts))

        raise TypeError(
            f"Invalid handler type: expected str, FunctionConfig, dict, or Function, "
            f"got {type(handler).__name__}"
        )

    @staticmethod
    def path_to_resource_name(path_parts: list[str]) -> str:
        """
        Convert path parts to a valid resource nam  e.
        Example: ['users', '{id}', 'orders'] -> 'users-id-orders'
        """
        # Remove any curly braces and convert to safe name
        safe_parts = [
            part.replace("{", "").replace("}", "").replace("+", "plus")
            for part in path_parts
        ]
        # TODO: check of longer than 256? if so cut the beginning or middle?
        return "-".join(safe_parts)

    @staticmethod
    def get_or_create_resource(
        path_parts: list[str], resources: dict[str, Resource], rest_api: RestApi
    ) -> Resource | None:
        if not path_parts:
            return None

        path_key = "/".join(path_parts)
        if path_key in resources:
            return resources[path_key]

        part = path_parts[-1]
        parent_parts = path_parts[:-1]
        parent_resource = (
            Api.get_or_create_resource(parent_parts, resources, rest_api)
            if parent_parts
            else None
        )
        parent_id = parent_resource.id if parent_resource else rest_api.root_resource_id
        resource = Resource(
            f"resource-{Api.path_to_resource_name(path_parts)}",
            rest_api=rest_api.id,
            parent_id=parent_id,
            path_part=part,
        )
        resources[path_key] = resource
        return resource

    def _create_resource(self) -> RestApi:
        # This is what needs to be done:
        #   1. create rest api
        #   2. for each route:
        #       a. create resource
        #       b. create method(s)
        #       c. create lambda from handler if it doesn't exists (we need to group
        #           routes based on lambda)
        #       d. create integration between method and lambda
        #       e. give lambda resource policy so it can be called by given
        #           gateway/resource/method
        #           https://docs.aws.amazon.com/lambda/latest/dg/access-control-resource-based.html
        #   3. create role for gateway and give it permission to write to cloudwatch
        #   4. create account and give it a role
        #   4. create deployment
        #   5. create stage
        # TODO:  we should validate conflicts in routes e.g. two same methods for same paths
        rest_api = RestApi(self.name)
        _create_api_gateway_account_and_role()

        resources = {}
        methods = []
        integrations = []
        grouped_routes_by_lambda = _group_routes_by_lambda(self._routes)
        group_config_map = _get_group_config_map(grouped_routes_by_lambda)

        # TODO: Refactor this section to use functional  patterns for better clarity
        #       Current implementation uses nested loops  but could be simplified using
        #       functional transformations or at least extrac different parts to separate
        #       functions.. This works reliably for now but will be optimized in future
        #       releases.
        for key, group in grouped_routes_by_lambda.items():
            route_with_config: _ApiRoute = group_config_map[key]
            routing_file_content = _create_routing_file(group, route_with_config)
            if isinstance(route_with_config.handler, Function):
                function = route_with_config.handler
            else:
                if isinstance(route_with_config.handler, FunctionConfig):
                    function_config = route_with_config.handler
                elif isinstance(route_with_config.handler, dict):
                    function_config = FunctionConfig(**route_with_config.handler)
                else:
                    raise ValueError(f"Bad type of route handler")
                extra_assets = {}
                if routing_file_content:
                    extra_assets["stlv_routing_handler.py"] = StringAsset(
                        routing_file_content
                    )

                # TODO: find better naming strategy, for now use key which is path to func and replace / with -
                #       this will not work if one function used by multiple APIs
                function = Function(key.replace("/", "-"), function_config)
                if extra_assets:
                    FunctionAssetsRegistry.add(function, extra_assets)

            # Randomly (like ever 10th time) pulumi fails to create this permission and as a consequence
            # an integration. No idea why yet. Happens indeterministically. Need to investigate.
            Permission(
                f"{self.name}-{function.name}-policy-statement",
                action="lambda:InvokeFunction",
                function=function.resource_name,
                principal="apigateway.amazonaws.com",
                source_arn=rest_api.execution_arn.apply(
                    lambda execution_arn: f"{execution_arn}/*"
                ),
            )

            for route in group:
                path_parts = [p for p in route.path.split("/") if p]
                resource = self.get_or_create_resource(path_parts, resources, rest_api)

                for http_method in route.methods:
                    method = Method(
                        f"method-{http_method}-{self.path_to_resource_name(path_parts)}",
                        rest_api=rest_api.id,
                        resource_id=resource.id,
                        http_method=http_method,
                        authorization="NONE",
                    )
                    methods.append(method)

                    integration = Integration(
                        f"integration-{http_method}-{route.path}",
                        rest_api=rest_api.id,
                        resource_id=resource.id,
                        http_method=method.http_method,
                        integration_http_method="POST",
                        type="AWS_PROXY",
                        uri=function.invoke_arn,
                    )
                    integrations.append(integration)
        stage = Stage(
            f"{self.name}-v1",
            rest_api=rest_api.id,
            deployment=_create_deployment(
                rest_api, self.name, methods + integrations
            ).id,
            stage_name="v1",
            # xray_tracing_enabled=True,
            access_log_settings={
                "destination_arn": rest_api.name.apply(
                    lambda name: f"arn:aws:logs:{get_region().name}:{get_caller_identity().account_id}:log-group:/aws/apigateway/{name}"
                ),
                "format": '{"requestId":"$context.requestId", "ip": "$context.identity.sourceIp", "caller":"$context.identity.caller", "user":"$context.identity.user","requestTime":"$context.requestTime", "httpMethod":"$context.httpMethod","resourcePath":"$context.resourcePath", "status":"$context.status","protocol":"$context.protocol", "responseLength":"$context.responseLength"}',
            },
            variables={"loggingLevel": "INFO"},
        )
        ComponentRegistry.add_instance_output(self, rest_api)
        pulumi.export(f"restapi_{self.name}_arn", rest_api.arn)
        pulumi.export(f"invoke_url_for_restapi_{self.name}", stage.invoke_url)

        return rest_api


def _create_deployment(api: RestApi, api_name: str, depends_on) -> Deployment:
    def calculate_trigger_hash(api_config):
        """
        Creates a hash of API configuration to determine if redeployment is needed.
        Changes to this hash will trigger a new deployment.
        """
        return sha256(json.dumps(api_config, sort_keys=True).encode()).hexdigest()

    # TODO: Implement proper configuration detection using api routes
    # Configuration that should trigger redeployment when changed
    api_config = {
        # "resources": {
        #     "paths": [],
        #     "methods": [],
        #     "integrations": []
        # },
        # Temporarily using timestamp to force deployment every time
        "timestamp": datetime.now().isoformat()
    }
    return Deployment(
        f"{api_name}-deployment",
        rest_api=api.id,
        # Trigger new deployment only when API config changes
        triggers={"redeployment": calculate_trigger_hash(api_config)},
        # Ensure deployment happens after all resources are created
        opts=ResourceOptions(depends_on=depends_on),
    )


@cache
def _create_api_gateway_account_and_role() -> Account:
    api_role = _create_api_gateway_role()
    return Account("api-gateway-account", cloudwatch_role_arn=api_role.arn)


API_GATEWAY_ROLE_NAME = "api-gateway-role"


@cache
def _create_api_gateway_role() -> Role:
    """Create basic execution role for API Gateway."""
    assume_role_policy = get_policy_document(
        statements=[
            GetPolicyDocumentStatementArgs(
                actions=["sts:AssumeRole"],
                principals=[
                    GetPolicyDocumentStatementPrincipalArgs(
                        identifiers=["apigateway.amazonaws.com"], type="Service"
                    )
                ],
            )
        ]
    )
    role = Role(API_GATEWAY_ROLE_NAME, assume_role_policy=assume_role_policy.json)

    RolePolicyAttachment(
        f"{API_GATEWAY_ROLE_NAME}-logs-policy-attachment",
        role=role.name,
        policy_arn=API_GATEWAY_LOGS_POLICY,
    )
    return role


def _group_routes_by_lambda(routes: list[_ApiRoute]) -> dict[str, list[_ApiRoute]]:
    def extract_key(handler_str: str) -> str:
        parts = handler_str.split("::")
        return parts[0] if len(parts) > 1 else handler_str.split(".")[0]

    grouped_routes = {}

    for route in routes:
        if isinstance(route.handler, Function):
            key = route.handler.name
        elif isinstance(route.handler, (FunctionConfig, dict)):
            config = (
                route.handler
                if isinstance(route.handler, FunctionConfig)
                else FunctionConfig(**route.handler)
            )
            key = config.folder if config.folder else extract_key(config.handler)
        else:
            key = extract_key(route.handler)

        grouped_routes.setdefault(key, []).append(route)

    return grouped_routes


def _get_group_config_map(
    grouped_routes: dict[str, list[_ApiRoute]]
) -> dict[str, _ApiRoute]:
    key_handler_config = {}
    for key, routes in grouped_routes.items():
        config_routes = []
        for route in routes:
            if isinstance(route.handler, (FunctionConfig, dict)):
                config = (
                    route.handler
                    if isinstance(route.handler, FunctionConfig)
                    else FunctionConfig(**route.handler)
                )
                # Check if there are any non-default configs besides handler/folder
                config_dict = {
                    k: v
                    for k, v in config.__dict__.items()
                    if k not in ("handler", "folder") and v
                }
                if any(v is not None for v in config_dict.values()):
                    config_routes.append(route)

        if len(config_routes) > 1:
            paths = [r.path for r in config_routes]
            raise ValueError(
                f"Multiple routes trying to configure the same lambda function: {', '.join(paths)}"
            )
        key_handler_config[key] = config_routes[0]

    return key_handler_config


def _create_route_map(routes: list[_ApiRoute]) -> dict[str, Tuple[str, str]]:
    route_map = {}
    for route in routes:
        for method in route.methods:
            key = f"{method} {route.path}"
            # we need to get file name and function name to import

            config: FunctionConfig
            if isinstance(route.handler, str):
                config = FunctionConfig(handler=route.handler)
            elif isinstance(route.handler, dict):
                config = FunctionConfig(**route.handler)
            else:
                config = route.handler

            route_map[key] = (
                config.local_handler_file_path,
                config.handler_function_name,
            )
    return route_map


def _create_routing_file(
    routes: list[_ApiRoute], config_route: _ApiRoute
) -> str | None:
    if isinstance(config_route.handler, Function) or len(routes) == 1:
        return None
    route_map = _create_route_map(routes)
    # If all roues points to same handler that means user is handling routing
    # so no need to generate the file
    if len(set(route_map.values())) < 2:
        return None
    return _generate_handler_file_content(route_map)


def _generate_handler_file_content(route_map: dict[str, Tuple[str, str]]) -> str:
    # Extract unique handler function names
    file_funcs = defaultdict(list)
    for file, func in route_map.values():
        file_funcs[file].append(func)

    # Generate imports section
    imports = [
        "# stlv_routing_handler.py",
        "# Auto-generated file - do not edit manually",
        "",
        "from typing import Any",
    ]

    for file, funcs in file_funcs.items():
        imports.append(f"from {file} import {', '.join(funcs)}")

    imports.extend(["", ""])

    # Generate routes dictionary
    routes_lines = ["ROUTES = {"]
    for route_key, func in route_map.items():
        routes_lines.append(f'    "{route_key}": {func[1]},')
    routes_lines.append("}")
    routes_lines.append("")
    routes_lines.append("")

    # Add the standard handler function
    handler_func = [
        "def lambda_handler(event: dict[str, Any], context: Any) -> dict[str, Any]:",
        '    method = event["httpMethod"]',
        '    resource = event["resource"]',
        '    route_key = f"{method} {resource}"',
        "",
        "    func = ROUTES.get(route_key)",
        "    if not func:",
        # TODO: add message with status code
        '        return {"statusCode": 500}',
        "    return func(event, context)",
        "",
    ]
    # Combine all sections
    content = imports + routes_lines + handler_func
    return "\n".join(content)



from enum import Enum
from typing import final

import pulumi
from pulumi import Output
from pulumi_aws.dynamodb import Table

from stelvio.aws.permission import AwsPermission
from stelvio.component import Component, ComponentRegistry, link_config_creator
from stelvio.link import Linkable, Link, LinkConfig


class AttributeType(Enum):
    STRING = "S"
    NUMBER = "N"
    BINARY = "B"


@final
class DynamoTable(Component[Table], Linkable):
    def __init__(
        self,
        name: str,
        *,
        fields: dict[str, AttributeType],
        partition_key: str,
        sort_key: str | None = None,
    ):
        super().__init__(name)
        # TODO:  links?
        self._fields = fields
        self._partition_key = partition_key
        self._sort_key = sort_key

    @property
    def fields(self) -> dict[str, AttributeType]:
        return dict(self._fields)  # Return a copy to prevent modification

    @property
    def partition_key(self) -> str:
        return self._partition_key

    @property
    def sort_key(self) -> str | None:
        return self._sort_key

    @property
    def arn(self) -> Output[str]:
        return self._resource.arn

    def _create_resource(self) -> Table:
        assert self._partition_key in self.fields, "partition_key not in fields list"
        if self._sort_key:
            assert self.sort_key in self.fields, "sort_key not in fields list"

        table = Table(
            self.name,
            billing_mode="PAY_PER_REQUEST",
            hash_key=self.partition_key,
            range_key=self.sort_key,
            attributes=[{"name": k, "type": v.value} for k, v in self.fields.items()],
        )
        pulumi.export(f"dynamo_{self.name}_arn", table.arn)
        return table

    # we can also provide other predefine links e.g read only, index etc. in addition to this default
    def link(self) -> Link:
        link_creator_ = ComponentRegistry.get_link_config_creator(type(self))

        link_config = link_creator_(self._resource)
        return Link(self.name, link_config.properties, link_config.permissions)


@link_config_creator(DynamoTable)
def default_dynamo_table_link(table: Table) -> LinkConfig:
    # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_examples_lambda-access-dynamodb.html
    return LinkConfig(
        properties={"table_arn": table.arn, "table_name": table.name},
        permissions=[
            AwsPermission(
                actions=[
                    "dynamodb:Scan",
                    "dynamodb:Query",
                    "dynamodb:GetItem",
                    "dynamodb:PutItem",
                    "dynamodb:UpdateItem",
                    "dynamodb:DeleteItem",
                ],
                resources=[table.arn],
            )
        ],
    )



from abc import ABC, abstractmethod
from functools import wraps
from typing import ClassVar, Callable, Any
from typing import Iterator

from pulumi import Resource as PulumiResource

from stelvio.link import LinkConfig


class Component[ResourceT: PulumiResource](ABC):
    # TODO: need to be unique, so validate for uniqueness
    _name: str

    def __init__(
        self,
        name: str,
    ) -> None:
        self._name = name
        print(f"{self._name}")
        ComponentRegistry.add_instance(self)

    @property
    def name(self) -> str:
        return self._name

    @property
    def _resource(self) -> ResourceT:
        if existing := ComponentRegistry.get_output(self):
            return existing

        resource = self._create_resource()
        ComponentRegistry.add_instance_output(self, resource)
        return resource

    def _ensure_resource(self) -> None:
        # Just triggers creation/ensures existence
        _ = self._resource

    @abstractmethod
    def _create_resource(self) -> ResourceT:
        """Implement actual resource creation logic"""
        ...


class ComponentRegistry:
    _instances: ClassVar[dict[type[Component], list[Component]]] = {}
    _instance_output_pairs: ClassVar[dict[Component, PulumiResource]] = {}
    _type_link_creators: ClassVar[dict[type, Callable]] = {}

    @classmethod
    def add_instance(cls, instance: Component[Any]):
        if type(instance) not in cls._instances:
            cls._instances[type(instance)] = []
        cls._instances[type(instance)].append(instance)

    @classmethod
    def add_instance_output[T: PulumiResource](cls, instance: Component[T], output: T):
        cls._instance_output_pairs[instance] = output

    @classmethod
    def get_output[T: PulumiResource](cls, instance: Component[T]) -> T:
        return cls._instance_output_pairs.get(instance)

    @classmethod
    def register_link_config_creator[
        T: PulumiResource
    ](cls, component_type: type[Component[T]], creator_fn: Callable[[T], LinkConfig]):
        cls._type_link_creators[component_type] = creator_fn

    @classmethod
    def get_link_config_creator[
        T: PulumiResource
    ](cls, component_type: type[Component[T]]) -> Callable[[T], LinkConfig] | None:
        return cls._type_link_creators.get(component_type)

    @classmethod
    def all_instances(cls) -> Iterator[Component[Any]]:
        instances = cls._instances.copy()
        for k in instances:
            for i in instances[k]:
                yield i


def link_config_creator[T: PulumiResource](component_type: type[Component[T]]):
    def decorator(func: Callable[[T], LinkConfig]) -> Callable[[T], LinkConfig]:
        @wraps(func)
        def wrapper(resource: T) -> LinkConfig:
            return func(resource)

        ComponentRegistry.register_link_config_creator(component_type, func)
        return wrapper

    return decorator



from dataclasses import dataclass

from pulumi import Input
from pulumi_aws.iam import GetPolicyDocumentStatementArgsDict


@dataclass(frozen=True)
class AwsPermission:
    actions: list[str] | Input[str]
    resources: list[Input[str]] | Input[str]

    def to_provider_format(self) -> GetPolicyDocumentStatementArgsDict:
        return GetPolicyDocumentStatementArgsDict(
            actions=self.actions, resources=self.resources
        )



import os
from dataclasses import field
from functools import cache
from pathlib import Path
from typing import Dict, Sequence, Any, TypedDict, Unpack, ClassVar
from typing import final

import pulumi
from pulumi import (
    AssetArchive,
    FileAsset,
    Output,
    StringAsset,
    Asset,
)
from pulumi_aws import lambda_, iam

from stelvio.component import ComponentRegistry, Component
from stelvio.link import Link, Linkable

DEFAULT_RUNTIME = "python3.12"

LAMBDA_EXCLUDED_FILES = ["stlv.py", ".DS_Store"]  # exact file matches
LAMBDA_EXCLUDED_DIRS = ["__pycache__"]  # exact directory matches
LAMBDA_EXCLUDED_EXTENSIONS = [".pyc"]  # file extensions


# "arn:aws:iam::aws:policy/service-role/AWSLambdaVPCAccessExecutionRole",
LAMBDA_BASIC_EXECUTION_ROLE = (
    "arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole"
)

from dataclasses import dataclass


@dataclass(frozen=True, kw_only=True)
class FunctionConfig:
    # handler is mandatory but rest defaults to None. Default values will be configured
    # elsewhere not here so they can be configurable and so they don't cause trouble
    # in api Gateway where we check for conflicting configurations
    handler: str
    folder: str | None = None
    links: list[Link | Linkable] = field(default_factory=list)
    memory_size: int | None = None
    timeout: int | None = None
    environment: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self):
        # Split handler by :: to check if using folder:handler format
        folder_parts = self.handler.split("::")

        if len(folder_parts) > 2:
            raise ValueError("Handler can only contain one :: separator")

        # If handler contains ::, validate folder and handler parts
        if len(folder_parts) == 2:
            if self.folder is not None:
                raise ValueError(
                    "Cannot specify folder both in handler (::) and as folder parameter"
                )
            folder_path, handler_path = folder_parts
            # Validate the extracted folder path
            if "." in folder_path:
                raise ValueError("Folder path should not contain dots")
        else:
            # If no :: in handler, validate any directly provided folder
            if self.folder is not None and "." in self.folder:
                raise ValueError("Folder path should not contain dots")
            handler_path = self.handler

        # Validate handler format: path/to/file.function_name
        if "." not in handler_path:
            raise ValueError(
                "Handler must contain a dot separator between file path and function "
                "name"
            )

        # Validate there's actual content before and after the dot
        file_path, function_name = handler_path.rsplit(".", 1)
        if not file_path or not function_name:
            raise ValueError("Both file path and function name must be non-empty")

        # Ensure file path doesn't contain any dots
        if "." in file_path:
            raise ValueError("File path part should not contain dots")

    @property
    def has_folder(self) -> bool:
        return self.folder is not None or "::" in self.handler

    @property
    def folder_path(self) -> str | None:
        if self.folder is not None:
            return self.folder
        return self.handler.split("::")[0] if self.has_folder else None

    @property
    def _handler_part(self) -> str:
        return self.handler.split("::")[1] if "::" in self.handler else self.handler

    @property
    def handler_file_path(self) -> str:
        # replace remaining dots with /  for non-root handler files?
        return self._handler_part.rsplit(".", 1)[0]

    @property
    def local_handler_file_path(self) -> str:
        # replace remaining dots with /  for non-root handler files?
        return self.handler_format.rsplit(".", 1)[0]

    @property
    def handler_function_name(self) -> str:
        return self._handler_part.rsplit(".", 1)[1]

    @property
    def handler_format(self) -> str:
        # If using folder, return just the handler part
        # For single file lambda, return everything after last slash
        return self._handler_part if self.has_folder else self.handler.split("/")[-1]


class FunctionConfigDict(TypedDict, total=False):
    handler: str
    folder: str
    links: list[Link | Linkable]
    memory_size: int
    timeout: int
    environment: dict[str, str]


# TODO: Think about what to make public interface/properties
@final
class Function(Component[lambda_.Function]):
    """AWS Lambda function component with automatic resource discovery.

    Generated environment variables follow pattern: STLV_RESOURCENAME_PROPERTYNAME

    Args:
        name: Function name
        config: Complete function configuration as FunctionConfig or dict
        **opts: Individual function configuration parameters

    You can configure the function in two ways:
        - Provide complete config:
            function = Function(
                name="process-user",
                config={"handler": "functions/orders.list", "timeout": 30}
            )
        - Provide individual parameters:
            function = Function(
                name="process-user",
                handler="functions/orders.list",
                links=[table.default_link(), bucket.readonly_link()]
            )
    """

    _config: FunctionConfig

    def __init__(
        self,
        name: str,
        config: None | FunctionConfig | FunctionConfigDict = None,
        **opts: Unpack[FunctionConfigDict],
    ):
        super().__init__(name)
        self._config = self._parse_config(config, opts)

    @staticmethod
    def _parse_config(
        config: None | FunctionConfig | FunctionConfigDict,
        opts: dict,
    ) -> FunctionConfig:
        if not config and not opts:
            raise ValueError(
                "Missing function handler: must provide either a complete "
                "configuration via 'config' parameter or at least the 'handler' option"
            )
        if config and opts:
            raise ValueError(
                "Invalid configuration: cannot combine 'config' parameter with "
                "additional options - provide all settings either in 'config' or as "
                "separate options"
            )

        if config is None:
            return FunctionConfig(**opts)
        if isinstance(config, FunctionConfig):
            return config
        if isinstance(config, dict):
            return FunctionConfig(**config)

        raise TypeError(
            "Invalid config type: expected FunctionConfig or dict, got "
            f"{type(config).__name__}"
        )

    @property
    def config(self) -> FunctionConfig:
        return self._config

    @property
    def invoke_arn(self) -> Output[str]:
        return self._resource.invoke_arn

    @property
    def resource_name(self) -> Output[str]:
        return self._resource.name

    def _create_resource(self) -> lambda_.Function:
        iam_statements = _extract_links_permissions(self._config.links)
        links_props = _extract_links_property_mappings(self._config.links)

        function_policy = _create_function_policy(self.name, iam_statements)
        lambda_role = _create_lambda_role(self.name)
        _attach_role_policies(self.name, lambda_role, function_policy)

        # https://docs.aws.amazon.com/lambda/latest/dg/configuration-envvars.html

        if self.config.has_folder:
            folder_path = self.config.folder_path
        else:
            folder_path = str(Path(self.config.handler_file_path).parent)

        lambda_resource_file_content = create_stlv_resource_file_content(links_props)

        LinkPropertiesRegistry.add(folder_path, links_props)

        ide_resource_file_content = create_stlv_resource_file_content(
            LinkPropertiesRegistry.get_link_properties_map(folder_path)
        )

        _create_stlv_resource_file(
            get_project_root() / folder_path, ide_resource_file_content
        )
        extra_assets_map = FunctionAssetsRegistry.get_assets_map(self)
        handler = self.config.handler_format
        if "stlv_routing_handler.py" in extra_assets_map:
            handler = "stlv_routing_handler.lambda_handler"
        function_resource = lambda_.Function(
            self.name,
            role=lambda_role.arn,
            runtime=DEFAULT_RUNTIME,
            code=_create_lambda_archive(
                self.config,
                lambda_resource_file_content,
                extra_assets_map,
            ),
            handler=handler,
            environment={"variables": _extract_links_env_vars(self._config.links)},
            memory_size=self.config.memory_size,
            timeout=self.config.timeout,
        )

        ComponentRegistry.add_instance_output(self, function_resource)
        pulumi.export(f"lambda_{self.name}_arn", function_resource.arn)
        return function_resource


class LinkPropertiesRegistry:
    _folder_links_properties_map: ClassVar[dict[str, dict[str, list[str]]]] = {}

    @classmethod
    def add(cls, folder: str, link_properties_map: dict[str, list[str]]):
        cls._folder_links_properties_map.setdefault(folder, {}).update(
            link_properties_map
        )

    @classmethod
    def get_link_properties_map(cls, folder: str) -> dict[str, list[str]]:
        return cls._folder_links_properties_map.get(folder, {})


class FunctionAssetsRegistry:
    _functions_assets_map: ClassVar[dict[Function, dict[str, Asset]]] = {}

    @classmethod
    def add(cls, function_: Function, assets_map: dict[str, Asset]):
        cls._functions_assets_map.setdefault(function_, {}).update(assets_map)

    @classmethod
    def get_assets_map(cls, function_: Function) -> dict[str, Asset]:
        return cls._functions_assets_map.get(function_, {}).copy()


def _extract_links_permissions(linkables: Sequence[Link | Linkable]) -> list[dict]:
    """Extracts IAM statements from permissions for function's IAM policy"""
    return [
        permission.to_provider_format()
        for linkable in linkables
        for permission in linkable.link().permissions
    ]


def _envar_name(link_name: str, prop_name: str) -> str:
    cleaned_link_name = "".join(c if c.isalnum() else "_" for c in link_name)

    if (first_char := cleaned_link_name[0]) and first_char.isdigit():
        cleaned_link_name = NUMBER_WORDS[int(first_char)] + cleaned_link_name[1:]

    return f"STLV_{cleaned_link_name.upper()}_{prop_name.upper()}"


def _extract_links_env_vars(linkables: Sequence[Link | Linkable]) -> dict[str, str]:
    """
    Creates environment variables with STLV_ prefix for runtime resource discovery.
    The STLV_ prefix in environment variables ensures no conflicts with other env vars
    and makes it clear which variables are managed by Stelvio.
    """
    link_objects = [item.link() for item in linkables]
    return {
        _envar_name(link.name, prop_name): value
        for link in link_objects
        for prop_name, value in link.properties.items()
    }


def _extract_links_property_mappings(
    linkables: Sequence[Link | Linkable],
) -> dict[str, list[str]]:
    """Maps resource properties to Python class names for code generation of resource access classes."""
    link_objects = [item.link() for item in linkables]
    return {link.name: [p for p in link.properties] for link in link_objects}


def _create_function_policy(
    name: str, statements: list[dict[str, Any]]
) -> iam.Policy | None:
    """Create IAM policy for Lambda if there are any statements."""
    if not statements:
        return None

    policy_document = iam.get_policy_document(statements=statements)
    return iam.Policy(
        f"{name}-Policy",
        name=f"{name}-Policy",
        path="/",
        policy=policy_document.json,
    )


def _create_lambda_role(name: str) -> iam.Role:
    """Create basic execution role for Lambda."""
    assume_role_policy = iam.get_policy_document(
        statements=[
            iam.GetPolicyDocumentStatementArgs(
                actions=["sts:AssumeRole"],
                principals=[
                    iam.GetPolicyDocumentStatementPrincipalArgs(
                        identifiers=["lambda.amazonaws.com"], type="Service"
                    )
                ],
            )
        ]
    )
    return iam.Role(f"{name}-Role", assume_role_policy=assume_role_policy.json)


def _attach_role_policies(
    name: str, role: iam.Role, function_policy: iam.Policy | None
) -> None:
    """Attach required policies to Lambda role."""
    iam.RolePolicyAttachment(
        f"{name}-BasicExecutionRolePolicyAttachment",
        role=role.name,
        policy_arn=LAMBDA_BASIC_EXECUTION_ROLE,
    )
    if function_policy:
        iam.RolePolicyAttachment(
            f"{name}-DefaultRolePolicyAttachment",
            role=role.name,
            policy_arn=function_policy.arn,
        )


def _create_lambda_archive(
    function_config: FunctionConfig,
    resource_file_content: str | None,
    extra_assets_map: dict[str, Asset],
) -> AssetArchive:
    """
    Create an AssetArchive for Lambda function based on configuration.
    Handles both single file and folder-based Lambdas.
    """
    project_root = get_project_root()
    assets = extra_assets_map
    handler_file = str(Path(function_config.handler_file_path).with_suffix(".py"))
    if function_config.has_folder:
        # Handle folder-based Lambda
        folder_path = project_root / function_config.folder_path
        if not folder_path.exists():
            raise ValueError(f"Folder not found: {folder_path}")

        # Check if handler file exists in the folder
        if handler_file not in extra_assets_map:
            absolute_handler_file = folder_path / handler_file
            if not absolute_handler_file.exists():
                raise ValueError(
                    f"Handler file not found in folder: {absolute_handler_file}.py"
                )

        # Recursively collect all files from the folder
        assets |= {
            str(file_path.relative_to(folder_path)): FileAsset(file_path)
            for file_path in folder_path.rglob("*")
            if not (
                file_path.is_dir()
                or file_path.name in LAMBDA_EXCLUDED_FILES
                or file_path.parent.name in LAMBDA_EXCLUDED_DIRS
                or file_path.suffix in LAMBDA_EXCLUDED_EXTENSIONS
            )
        }
    else:
        # Handle single file Lambda
        if handler_file not in extra_assets_map:
            absolute_handler_file = project_root / handler_file
            if not absolute_handler_file.exists():
                raise ValueError(f"Handler file not found: {absolute_handler_file}")
            assets[absolute_handler_file.name] = FileAsset(absolute_handler_file)
    if resource_file_content:
        assets["stlv_resources.py"] = StringAsset(resource_file_content)
    return AssetArchive(assets)


def _create_stlv_resource_file(folder: Path, content: str) -> None:
    """Create resource access file with supplied content."""
    path = folder / "stlv_resources.py"

    # Delete file if no content
    if not content:
        path.unlink(missing_ok=True)
        return

    with open(path, "w") as f:
        f.write(content)


def _single_file_create_stlv_resource_file(
    folder: Path, link_properties_map: Dict[str, list[str]]
) -> None:
    """Generate resource access file with classes for linked resources."""
    path = folder / "stlv_resources.py"

    # Delete file if no properties to generate
    if not any(link_properties_map.values()):
        path.unlink(missing_ok=True)
        return
    with open(path, "w") as f:
        f.write("import os\n")
        f.write("from dataclasses import dataclass\n")
        f.write("from typing import Final\n")
        f.write("from functools import cached_property\n\n\n")

        for link_name, properties in link_properties_map.items():
            if not properties:
                continue
            class_name = _to_valid_python_class_name(link_name)
            f.write("@dataclass(frozen=True)\n")
            f.write(f"class {class_name}Resource:\n")
            for prop in properties:
                f.write(f"    @cached_property\n    def {prop}(self) -> str:\n")
                f.write(
                    f'        return os.getenv("{_envar_name(link_name, prop)}")\n\n'
                )
            f.write("\n")

        f.write("@dataclass(frozen=True)\n")
        f.write("class LinkedResources:\n")
        for link_name in link_properties_map:
            class_name = _to_valid_python_class_name(link_name)
            f.write(
                f"    {_pascal_to_camel(class_name)}: Final[{class_name}Resource] = {class_name}Resource()\n"
            )
        f.write("\n\n")

        f.write("Resources: Final = LinkedResources()\n")


def create_stlv_resource_file_content(
    link_properties_map: Dict[str, list[str]]
) -> str | None:
    """Generate resource access file content with classes for linked resources."""
    # Return None if no properties to generate
    if not any(link_properties_map.values()):
        return None

    lines = [
        "import os",
        "from dataclasses import dataclass",
        "from typing import Final",
        "from functools import cached_property\n\n",
    ]

    for link_name, properties in link_properties_map.items():
        if not properties:
            continue
        lines.extend(_create_link_resource_class(link_name, properties))

    lines.extend(["@dataclass(frozen=True)", "class LinkedResources:"])

    # and this
    for link_name in link_properties_map:
        class_name = _to_valid_python_class_name(link_name)
        lines.append(
            f"    {_pascal_to_camel(class_name)}: Final[{class_name}Resource] = {class_name}Resource()"
        )
    lines.extend(["\n", "Resources: Final = LinkedResources()"])

    return "\n".join(lines)


def _create_link_resource_class(link_name, properties) -> list[str] | None:
    if not properties:
        return None
    lines = []
    class_name = _to_valid_python_class_name(link_name)
    lines.append("@dataclass(frozen=True)")
    lines.append(f"class {class_name}Resource:")
    for prop in properties:
        lines.extend(
            [
                f"    @cached_property",
                f"    def {prop}(self) -> str:",
                f'        return os.getenv("{_envar_name(link_name, prop)}")\n',
            ]
        )
    lines.append("")
    return lines


NUMBER_WORDS = {
    "0": "Zero",
    "1": "One",
    "2": "Two",
    "3": "Three",
    "4": "Four",
    "5": "Five",
    "6": "Six",
    "7": "Seven",
    "8": "Eight",
    "9": "Nine",
}


def _to_valid_python_class_name(aws_name: str) -> str:
    # Split and clean
    words = aws_name.replace("-", " ").replace(".", " ").replace("_", " ").split()
    cleaned_words = ["".join(c for c in word if c.isalnum()) for word in words]
    class_name = "".join(word.capitalize() for word in cleaned_words)

    # Convert only first digit if name starts with number
    if class_name and class_name[0].isdigit():
        first_digit = NUMBER_WORDS[class_name[0]]
        class_name = first_digit + class_name[1:]

    return class_name


def _pascal_to_camel(pascal_str: str) -> str:
    """Convert Pascal case to camel case.
    Example: PascalCase -> pascalCase, XMLParser -> xmlParser
    """
    if not pascal_str:
        return pascal_str
    i = 1
    while i < len(pascal_str) and pascal_str[i].isupper():
        i += 1
    return pascal_str[:i].lower() + pascal_str[i:]


# It's here for now but probably will be moved to some other place as it's used in
# other places too
@cache
def get_project_root() -> Path:
    """
    Find and cache the project root by looking for stlv_app.py.
    Raises ValueError if not found.
    """
    start_path = Path(os.getcwd()).resolve()

    current = start_path
    while current != current.parent:
        if (current / "stlv_app.py").exists():
            return current
        current = current.parent

    raise ValueError(
        "Could not find project root: no stlv_app.py found in parent directories"
    )
